/*
 *	$Id: casdef.h,v 1.1 1995/08/12 00:13:33 jhill Exp $
 *
 *      Author: Jeffrey O. Hill
 *              johill@lanl.gov
 *              (505) 665 1831
 *      Date:   1-95
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 *      Modification Log:
 *      -----------------
 *	$Log: casdef.h,v $
 * Revision 1.1  1995/08/12  00:13:33  jhill
 * installed into cvs
 *
 */

/*
 * to do
 *
 * 1) how do they delete a pv out from under CA
 *
 */

#include <epicsTypes.h>	/* EPICS arch independent types */
#include <tsDefs.h>	/* EPICS time stamp */
#include <alarm.h>	/* EPICS alarm severity and alarm condition */
#include <errMdef.h>	/* EPICS error codes */

typedef int		caStatus;

typedef unsigned	caid;
typedef caid		chId;
typedef caid		pvId;

typedef struct{
	unsigned long	sec;	/* seconds */
	unsigned long	nsec;	/* nano - seconds */
}caTime;


/*
 * ===========================================================
 * provided by the application (and called by the server via 
 * a jump table)
 *
 * All of these operations need to return without delay
 * if we are to avoid hanging a single threaded server.
 * ===========================================================
 */

/*
 * return true if the named PV exists and false if it does not
 *
 * Canonical name is copied by the application into the buf pointed to by 
 * officialPVNameBuf. Copy no more than officialPVNameBufSize
 * characters and null terminate. 
 *
 * This is intended to support PV name aliases.
 */
typedef epicsBoolean pvExistTest (const char *pPVName, char *officialPVNameBuf, 
			unsigned officialPVNameBufSize);

/* 
 * The application supplied entry point pvAddrCreate() will
 * be called by the server library each time that
 * CA attaches a channel to a process varible for 
 * the first time.
 *
 * The application typically will allocate an application specific 
 * address structure with caMalloc() and return a pointer to it 
 * in *ppPVAddr. 
 */
typedef caStatus pvAddrCreate (const char *pvNameString, 
				pvId id, void **ppPVAddr);

/* 
 * The application supplied entry point pvAddrDelete() will
 * be called by the server library each time that
 * the number of CA attachments to a particular PV
 * decrements to zero. This routine provide an opportunity 
 * for the application to delete any resources allocated 
 * during pvAddrCreate(). Typically caFree() will be called
 * to free memmory allocated inside of pvAddrCreate() above.
 */
typedef caStatus pvAddrDelete (void *pPVAddr);



/* 
 * The application supplied entry point pvInterestRegister ()
 * will be called each time that the server wishes to 
 * subscripe for PV value change events via caServerPostEvents() 
 */
typedef caStatus pvInterestRegister (void *pPVAddr, unsigned select);

/* 
 * The application supplied entry point pvInterestDelete ()
 * will be called each time that the server wishes to
 * remove its subscription for PV value change events 
 * via caServerPostEvents() 
 */
typedef caStatus pvInterestDelete (void *pPVAddr);




/*
 * No changes to a PV while the lock is applied
 *
 * allows updates to a record to be postponed while
 * options are being read so that the value and options 
 * are consistent.
 */
typedef caStatus pvLock (void *pPVAddr);
typedef caStatus pvUnlock (void *pPVAddr);


/* 
 * Always called just after any write to the native 
 * address (with the lock on) 
 *
 * If this operation will not complete immediately
 * then the application is required to return
 * S_casApp_asyncCompletion and then call 
 * caServerPostPVIOCompletion ()
 * when the operation actually completes.
 * The server will allow only one asynchronous IO
 * operation at a time to be pending on a PV.
 */
typedef caStatus pvWriteAction (void *pPVAddr);

/* 
 * always called just prior to a read from the native 
 * address (with the lock on) 
 *
 * If this operation will not complete immediately
 * then the application is required to return
 * S_casApp_asyncCompletion and then call 
 * caServerPostPVIOCompletion ()
 * when the operation actually completes.
 * The server will allow only one asynchronous IO
 * operation at a time to be pending on a PV.
 */
typedef caStatus pvReadAction (void *pPVAddr);

/* 
 * allowed to change during an id's lifetime ??
 */
typedef caStatus pvNativeType (void *pPVAddr, epicsType *pType);
typedef caStatus pvNativeElementCount (void *pPVAddr, unsigned long *pCount);
typedef caStatus pvNativeAddress (void *pPVAddr, void **pPointer);
/* ring index certainly will chnage during the PV's lifespan) */
typedef caStatus pvNativeRingIndex (void *pPVAddr, unsigned long *pIndex);

/*
 * frequently change during an id's lifetime
 */
typedef caStatus pvAlarmStatus (void *pPVAddr, epicsAlarmSeverity *pSeverity,
				epicsAlarmCondition *pStatus);
typedef caStatus pvTimeStamp (void *pPVAddr, TS_STAMP *pTimeStamp);

/*
 * Will cause operational problems for existing clients coded to the
 * current interface if these change during an id's lifetime. Future 
 * client interface will provide for change notification. Perhaps
 * we should just disconnect the client if these parameters change.
 */
typedef	caStatus pvBestExternalType (void *pPVAddr, epicsType *pType);	
typedef caStatus pvUnits (void *pPVAddr, unsigned bufByteSize, 
				char *pStringBuf);
typedef caStatus pvEnumStateCount (void *pPVAddr, unsigned nStatesMax,
					unsigned *nStates);
typedef caStatus pvEnumStateString (void *pPVAddr, unsigned stateNumber,
				unsigned bufByteSize, char *pStringBuf);
typedef caStatus pvPrecision (void *pPVAddr, unsigned *pPrecision);
typedef caStatus pvGraphLimits (void *pPVAddr, double *pUpperLimit, 
			double *pLowerLimit);
typedef caStatus pvControlLimits (void *pPVAddr, double *pUpperLimit, 
			double *pLowerLimit);
typedef caStatus pvAlarmLimits (void *pPVAddr, double *pUpperAlarmLimit, 
			double *pUpperWarningLimit, double *pLowerWarningLimit, 
			double *pLowerAlarmLimit);

#if 0 /* new stuf */

typedef struct cas_value_log {
	epicsAny		value;
	epicsAlarmSeverity	severity;
	epicsAlarmCondition	condition;
	TS_STAMP		time;
}casValueLog;

/*
 * ndim =0 => scaler 
 * otherwise pIndexArray points to an array of ndim items
 */
typedef struct cas_io_context {
	void 		*pPVAddr;
	epicsArrayIndex	*pIndexArray;	/* pointer to index array */
	unsigned	ndim;		/* number of dimensions */
	casIOID		ioid;		/* for async completion */
}casCtx;

typedef caStatus pvWrite (const casCtx *pCtx, const epicsAny *pValue);
/*
 * If it is an array operation then the array entry in epicsAny
 * in casValueLog is initialized (including all pointers) so that
 * the application can return the value directly into an
 * array allocated by the server library
 */
typedef caStatus pvRead (const casCtx *pCtx, casValueLog *pLog);

typedef enum {
	casevExist,		/* exist test async completion */
	casevCreate,		/* PV create async completion */
	casevWrite,		/* async write completion */
	casevRead,		/* async read completion */
	casevDelete,		/* PV delete (initiated by app) */ 
	casevMonitor		/* interest call back */
	} casEventType;

typedef struct {
	caStatus	status;
	char		*pOfficialName;
}casevExist;

typedef struct {
	caStatus	status;
	void		*pPVAddr;
}casevCreate;

typedef struct {
	caStatus	status;
	casIOID		ioid;	
}casevWrite;

typedef struct {
	caStatus	status;
	casIOID		ioid;	
	casValueLog	log;
}casevRead;

typedef struct {
	pvId		id;
	casValueLog	log;
	unsigned	select;
}casevMonitor;

typedef struct {
}casevDelete;

typedef union {
	casEventType	type;
	TS_STAMP	time;
	union {
		casevExist	exist;
		casevCreate	create;
		casevRead	read;
		casevWrite	write;
		casevMonitor	monitor;
		casevDelete 	delete;
	}detail;
}casEvent;

caStatus caServerPostEvent (caServerId casid, casEvent *pEvent)

#endif  /* new stuf */

/* 
 * Optional per channel interface
 * 
 * Per channel state required if the application
 * implements access control
 */
typedef caStatus chAddrCreate (void *pPVAddr, chId id, void **ppChanAddr);
typedef caStatus chAddrDelete (void *pChAddr);
typedef caStatus chSetOwner (void *pChAddr, const char *pUser, const char *pHost);
#define channelInterest_AccessRights (1<<0)
typedef caStatus chInterestRegister (void *pChAddr, unsigned select);
typedef caStatus chInterestDelete (void *pChAddr);
typedef epicsBoolean chReadAccess (void *pChAddr);
typedef epicsBoolean chWriteAccess (void *pChAddr);

/*
 * If an entry point isnt supplied by the application
 * (set to NULL) then the server lib will fill in 
 * a default action
 */
typedef struct cas_application_entry_table{
	pvExistTest		*pPVExistTest;
	pvAddrCreate		*pPVAddrCreate;
	pvAddrDelete		*pPVAddrDelete;

	pvInterestRegister	*pPVInterestRegister;
	pvInterestDelete	*pPVInterestDelete;

	pvLock			*pPVLock;
	pvUnlock		*pPVUnlock;

	pvWriteAction		*pPVWriteAction;
	pvReadAction		*pPVReadAction;

	pvAlarmStatus		*pPVAlarmStatus;
	pvTimeStamp		*pPVTimeStamp;

	pvNativeType		*pPVNativeType;
	pvBestExternalType	*pPVBestExternalType;
	pvNativeElementCount	*pPVNativeElementCount;
	pvNativeAddress		*pPVNativeAddress;
	pvNativeRingIndex	*pPVNativeRingIndex;
	pvUnits			*pPVUnits;
	pvPrecision		*pPVPrecision;
	pvEnumStateCount	*pPVEnumStateCount;
	pvEnumStateString	*pPVEnumStateString;
	pvGraphLimits		*pPVGraphLimits;
	pvControlLimits		*pPVControlLimits;
	pvAlarmLimits		*pPVAlarmLimits;

	chAddrCreate		*pChAddrCreate;
	chAddrDelete		*pChAddrDelete;
	chSetOwner		*pChSetOwner;
	chInterestRegister	*pChInterestRegister;
	chInterestDelete	*pChInterestDelete;
	chReadAccess		*pChReadAccess;
	chWriteAccess		*pChWriteAccess;
}casAppEntryTable;

/*
 * ===========================================================
 * called by the application
 * ===========================================================
 */

/*
 * Context used when creating a server.
 * Set appropriate flag in the "flags"
 * field for each parameter that does
 * not default.
 */
#define casCreateCtx_pvCountEstimate	(1<<1)
#define casCreateCtx_pvMaxNameLength	(2<<1) 	/* required */

typedef struct {
	unsigned 	flags;
	unsigned	pvCountEstimate; /* estimated max PV's in this server*/
	unsigned	pvMaxNameLength; /* ma char in PV name - required */
}casCreateCtx;

void 		*caMalloc (size_t size);
void 		*caCalloc (size_t count, size_t size);
void 		caFree (void *pBlock);

/*
 * NOTE: Always force casAppEntryTable to zero prior to initializatiion
 * so that if new entries are added to the bottom of the structure 
 * existing code will specify nill entries for any new entry points
 * and therefore take the default.
 */
typedef void 		*caServerId;
caStatus 	caServerCreate (const casAppEntryTable *pTable, 
			const casCreateCtx *pCtx, caServerId *pId);
caStatus 	caServerDelete (caServerId id);
caStatus	caServerProcess (caServerId id, const caTime *pDelay);
typedef unsigned 	caServerTimerId;
caStatus	caServerAddTimeout (caServerId id, const caTime *pDelay,
			void (*pFunc)(void *pParam), void *pParam, 
			caServerTimerId *pAlarmId);
caStatus	caServerDeleteTimeout (caServerId id, caServerTimerId alarmId);

caStatus 	caServerSetDebugLevel (caServerId id, unsigned level);
caStatus	caServerShow (caServerId id, unsigned level);	

/*
 * Application calls this function if a PV's state is modified.
 */
caStatus 	caServerPostPVEvent (caServerId casid, pvId id, 
			unsigned eventSelect);

/*
 * Application calls this function when an asynchronous IO operation
 * completes.  
 */
caStatus 	caServerPostPVIOCompletion (caServerId casid, 
			pvId id, caStatus status);

/*
 * Application calls this function if a channel's state is modified.
 * For example, the access rights change for a particular client attached 
 * to a particular PV may change.
 */
caStatus	caServerPostChEvent (caServerId casid, chId id, 
			unsigned eventSelect);

/*
 * ===========================================================
 * for internal use by the server library 
 * (and potentially returned to the server application)
 * ===========================================================
 */
#define S_cas_success 0
#define S_cas_internal (M_cas| 1) /*Internal failure*/
#define S_cas_noMemory (M_cas| 2) /*Memory allocation failed*/
#define S_cas_portInUse (M_cas| 3) /*IP port already in use*/
#define S_cas_hugeRequest (M_cas | 4) /*Requested op does not fit*/
#define S_cas_sendBlocked (M_cas | 5) /*Blocked for send q space*/
#define S_cas_badElementCount (M_cas | 6) /*Bad element count*/
#define S_cas_noConvert (M_cas | 7) /*No conversion between src & dest types*/
#define S_cas_badWriteType (M_cas | 8) /*Src type inappropriate for write*/
#define S_cas_ioBlocked (M_cas | 9) /*Blocked for io completion*/
#define S_cas_partialMessage (M_cas | 10) /*Partial message*/
#define S_cas_noContext (M_cas | 11) /*Context parameter is required*/
#define S_cas_disconnect (M_cas | 12) /*Lost connection to server*/
#define S_cas_recvBlocked (M_cas | 13) /*Recv blocked*/
#define S_cas_badType (M_cas | 14) /*Bad data type*/
#define S_cas_timerDoesNotExist (M_cas | 15) /*Timer does not exist*/


/*
 * ===========================================================
 * returned by the application (to the server library)
 * ===========================================================
 */
#define S_casApp_success 0 
#define S_casApp_noMemory (M_casApp | 1) /*Memory allocation failed*/
#define S_casApp_pvNotFound (M_casApp | 2) /*PV not found*/
#define S_casApp_badPVId (M_casApp | 3) /*Unknown PV identifier*/
#define S_casApp_noSupport (M_casApp | 4) /*No application support for op*/
#define S_casApp_asyncCompletion (M_casApp | 5) /*Operation will complete asynchronously*/

